module ST1Monad(
 ST, Ref, newRef, readRef, writeRef, runST
 ) where

import qualified Data.Map as M
import qualified STMonad as State hiding (cntToL) 
import Data.Char(isUpper, toLower)

-- Our state: a map from adresses to arbitrary values
newtype Ref alpha = Ref { addr :: Integer } deriving (Eq, Ord)
type Mem alpha = M.Map Integer alpha

newRef' :: Mem alpha-> Ref alpha
newRef' s = Ref $ maximum (0 : M.keys s)+ 1

readRef'  :: Ref alpha -> Mem alpha-> alpha
readRef' r m = m M.! (addr r)

writeRef' :: Ref alpha-> alpha-> Mem alpha-> Mem alpha
writeRef' r = M.insert (addr r)

-- The state transformer:
newtype ST alpha beta = ST { state :: State.ST (Mem alpha) beta }

-- runST :: (forall sigma. State sigma alpha) -> alpha
runST :: ST alpha beta-> beta
runST s = fst (State.run (state s) M.empty)

-- wrapping up our basic operations:
newRef :: alpha-> ST alpha (Ref alpha)
newRef a =  ST $ do r<- State.get newRef'
                    State.set $ writeRef' r a 
                    return r 

readRef :: Ref alpha -> ST alpha alpha
readRef a = ST $ State.get (readRef' a) 

writeRef :: Ref alpha -> alpha -> ST alpha ()
writeRef a v = ST $ State.set (writeRef' a v)

-- Unfortunately, we have to redo the type instances, but these
-- are largely schematic: unwrap, apply functions on State type, rewrap

instance Functor (ST alpha) where
  fmap f s = ST $ fmap f (state s)

instance Applicative (ST alpha) where
  pure a = ST $ pure a
  p <*> q = ST $ state p <*> state q

instance Monad (ST alpha) where
  p >>= q = ST $ state p >>= state. q 
  return a = ST $ return a
