module ST2Monad(
 ST, Ref, newRef, readRef, writeRef, runST
 ) where

import qualified Data.Map as M
import Data.Dynamic
import Data.Maybe

import qualified STMonad as State hiding (cntToL)
import Data.Char(isUpper, toLower)

-- Our state: a map from adresses to arbitrary values, wrapped up in 
-- a dynamic type.
newtype Ref alpha = Ref { addr :: Integer } deriving (Eq, Ord)
type Mem = M.Map Integer Dynamic

newRef' :: Mem -> Ref alpha
newRef' s = Ref $ maximum (0: M.keys s)+ 1

readRef'  :: Typeable alpha=> Ref alpha -> Mem-> alpha
readRef' a m = fromJust $ fromDynamic (m M.! (addr a))

writeRef' :: Typeable alpha=> Ref alpha-> alpha-> Mem-> Mem
writeRef' a v = M.insert (addr a) (toDyn v) 

-- The state transformer type:
newtype ST beta = ST { state :: State.ST Mem beta }

-- runST :: (forall sigma. State sigma alpha) -> alpha
runST :: ST beta-> beta
runST s = fst (State.run (state s) M.empty)

-- wrapping up our basic operations:
newRef :: Typeable alpha=> alpha-> ST (Ref alpha)
newRef v =  ST $ do r<- State.get newRef' 
                    State.set (writeRef' r v)
                    return r

readRef :: Typeable alpha=> Ref alpha -> ST alpha
readRef a = ST $ State.get (readRef' a) 

writeRef :: Typeable alpha=> Ref alpha -> alpha -> ST ()
writeRef a v = ST $ State.set (writeRef' a v)

-- Type instances

instance Functor ST where
  fmap f s = ST $ fmap f (state s)

instance Applicative ST where
  pure a = ST $ pure a
  p <*> q = ST $ state p <*> state q

instance Monad ST where
  p >>= q = ST $ state p >>= state. q 
  return a = ST $ return a


