module STMonad where

import Data.Char (isUpper, toLower)


-- Type representing basic stateful functions
-- We need to wrap this in a datatype to be able to make it into an instance
data ST sigma alpha = St { run :: sigma-> (alpha, sigma) }

-- Functor instance: lifting functions
instance Functor (ST sigma) where
  fmap f g = St $ \s-> let (a, s1)= run g s in (f a, s1)

-- Applicate instance: threading values
instance Applicative (ST sigma) where
  pure a = St $ \s-> (a, s)
  p <*> q = St $ \s-> let (f, s') = run p s
                          (a, s'') = run q s'
                      in  (f a, s'')

-- Monad instance: composition and lifiting
instance Monad (ST sigma) where
  f >>= g = St $ \s -> let (a, s')= run f s in  run (g a) s'
  return a = St $ \s-> (a, s)




-- Basic operations

-- Reading from the state
get :: (sigma-> alpha)-> ST sigma alpha
get f = St $ \s-> (f s, s)

-- Setting the state
set :: (sigma-> sigma)-> ST sigma ()
set g = St $ \s-> ((), g s)

-- Example using a very simple state
type WithCounter alpha = ST Int alpha

-- This function is like toLower, but also counts the number of upper-case
-- characters converted to lower.
cntToLower :: String-> (String, Int)
cntToLower s = run (cntToL s) 0

-- Main function : we can now use monad syntax!
cntToL :: String-> WithCounter String
cntToL [] = return ""
cntToL (x:xs) 
  | isUpper x =  do ys<- cntToL xs 
                    set (+1)
                    return (toLower x: ys)
  | otherwise =  do { ys<- cntToL xs; return (x: ys) }

